/** @file gui_base.h

	Skelton for retropc emulator
	SDL edition

	@author Sasaji
	@date   2015.04.01

	@brief [ gui base ]
*/

#ifndef _GUI_BASE_H_
#define _GUI_BASE_H_

#include <SDL.h>
#include <SDL_syswm.h>
#include "../common.h"
#include "../emu.h"
#include "../vm/vm.h"

class EMU;
class LedBox;

/// GUI wrapper class
class GUI_BASE
{
protected:
	EMU *emu;
	SDL_SysWMinfo info;
	int need_update_screen;
	bool globalkey_enable;

	LedBox *ledbox;

public:
	GUI_BASE(int argc, char **argv, EMU *new_emu);
	virtual ~GUI_BASE();

	/// @name send to gui from emu class or main.
	//@{

	/// @return -1:error 0:normal ok 1:ok but use default locale
	virtual int Init();
#ifdef GUI_USE_FOREIGN_WINDOW
	/// create window without SDL library
	virtual void *CreateWindow(int width, int height);
	virtual void *GetWindowData();
#endif
	/// create widget if need
	virtual int CreateWidget(SDL_Surface *screen, int width, int height);
	/// create menu if need
	/// @return -1:error 0:normal ok 1:ok (need window resize)
	virtual int CreateMenu();
	/// show gui menu
	virtual void ShowMenu();
	/// hide gui menu
	virtual void HideMenu();

	/// draw gui widget on surface
	/// @return 1:processed SDL_UpdateRect() in this function
	virtual int MixSurface();

	/// post update screen request message
	/// @attention this function is called by emu thread
	/// @return false:did not call UpdateScreen in main thread
	virtual bool NeedUpdateScreen();
	/// post finished updateing screen request message
	/// @attention this function is called by emu thread
	virtual void UpdatedScreen();
	/// draw screen on window
	/// @note this function should be called by main thread
	virtual void UpdateScreen();
	virtual void DecreaseUpdateScreenCount();

	virtual int CreateGlobalKeys();

	/// @return true : processed global key
	virtual bool ExecuteGlobalKeys(int code, uint32 status);
	/// @return true : processed global key
	virtual bool ReleaseGlobalKeys(int code, uint32 status);

	/// something do before process event
	virtual void PreProcessEvent();
	/// event processing
	/// @return -1: exit event loop 0: processed myself 1: no processing
	virtual int ProcessEvent(SDL_Event *e);
	/// something do after process event
	virtual void PostProcessEvent();

	/// callback from emu screen change
	virtual void ScreenModeChanged(bool fullscreen);

	/// post to SDL event message
	void PostCommandMessage(int id, void *data1 = NULL, void *data2 = NULL);

	///
	virtual void KeyDown(int code, uint32 mod);
	///
	virtual void KeyUp(int code);

	///
	virtual void SetFocusToMainWindow();

	/// ledbox
	virtual void CreateLedBox(const char *res_path, SDL_PixelFormat *src_format);
	virtual void CreateLedBoxSub();
	virtual void ReleaseLedBox();
	virtual void SetLedBoxPosition(bool mode, int left, int top, int width, int height, int place);
	virtual void UpdateLedBox(uint64 flag);
	virtual void DrawLedBox(SDL_Surface *screen);
	virtual void MoveLedBox();

	//@}

	/// @name send to ui class from ui (user / operator).
	//@{
	virtual void Exit(void);

	// Show Dialog

	virtual bool ShowLoadDataRecDialog(void);
	virtual bool ShowSaveDataRecDialog(void);

	virtual bool ShowOpenFloppyDiskDialog(int drv);
	virtual int  ShowSelectFloppyDriveDialog(int drv);

	virtual bool ShowLoadStateDialog(void);
	virtual bool ShowSaveStateDialog(bool cont);

	virtual bool ShowOpenAutoKeyDialog(void);

	virtual bool ShowPlayRecKeyDialog(void);
	virtual bool ShowRecordRecKeyDialog(void);

	virtual bool ShowRecordStateAndRecKeyDialog(void);

	virtual bool ShowSavePrinterDialog(int drv);

	virtual bool ShowRecordVideoDialog(int fps_num);
	virtual bool ShowRecordAudioDialog(void);
	virtual bool ShowRecordVideoAndAudioDialog(int fps_num);

	virtual bool ShowVolumeDialog(void);
	virtual bool ShowSndFilterDialog(void);

	virtual bool ShowKeybindDialog(void);
	virtual bool ShowConfigureDialog(void);

	virtual bool ShowAboutDialog(void);
	//@}

	/// @name callback to emu class from ui (user / operator).
	//@{

	// Control

	virtual void PostEtReset(void);
	virtual void PostEtSpecialReset(void);
	virtual void PostEtWarmReset(int onoff);
	virtual void Dipswitch(int bit);
	virtual uint8 GetDipswitch(void);
	virtual void PostEtTogglePause(void);
	virtual void TogglePause(void);
	virtual bool NowPause(void);
	virtual void PostEtSystemPause(bool val);
	virtual void SystemPause(bool val);
	virtual bool NowSystemPause(void);
	virtual void PostEtCPUPower(int num);
	virtual int  GetCPUPower(void);

	virtual void ChangeFddType(int num);
	virtual int  GetFddType(void);
	virtual int  NextFddType(void);
	virtual void PostEtToggleSyncIRQ(void);
	virtual bool NowSyncIRQ(void);

	// Screen

	virtual void ChangeFrameRate(int num);
	virtual int  GetFrameRateNum(void);
	virtual void TogglgScreenMode(void);
	virtual void ChangeWindowMode(int num);
	virtual void ChangeFullScreenMode(int num);
	virtual int  GetWindowMode(void);
	virtual int  GetFullScreenMode(void);
	virtual int  GetWindowModeCount(void);
	virtual bool GetWindowModeStr(int num, _TCHAR *str);
	virtual int  GetFullScreenModeCount(void);
	virtual bool GetFullScreenModeStr(int num, _TCHAR *str);
	virtual bool IsFullScreen(void);
	virtual void PostEtChangeScanLine(int num);
	virtual void PostEtChangeScanLine(void);
	virtual int  NowScanLine(void);
	virtual void ChangeStretchScreen(int num);
	virtual int  GetStretchScreen(void);
//	virtual void ToggleCutoutScreen(void);
//	virtual bool NowCutoutedScreen(void);
	virtual void ChangePixelAspect(int num);
	virtual int  GetPixelAspectMode(void);
	virtual int  GetPixelAspectModeCount(void);
	virtual void GetPixelAspectModeStr(int num, _TCHAR *str);

	virtual void PostEtStartRecordVideo(int num);
	virtual void PostEtStopRecordVideo(void);
	virtual bool NowRecordingVideo(void);
	virtual int  GetRecordVideoFrameNum(void);
	virtual void PostEtResizeRecordVideoSurface(int num);
	virtual int  GetRecordVideoSurfaceNum(void);
	virtual bool GetRecordVideoSizeStr(int num, _TCHAR *str);
	virtual void PostEtCaptureScreen(void);

	virtual void PostEtChangeAfterImage(int num);
	virtual void PostEtChangeAfterImage(void);
	virtual int  GetAfterImageMode(void);
	virtual void PostEtChangeKeepImage(int num);
	virtual void PostEtChangeKeepImage(void);
	virtual int  GetKeepImageMode(void);

#ifdef USE_OPENGL
	virtual void ChangeUseOpenGL(int num);
	virtual void ChangeUseOpenGL(void);
	virtual int  GetOpenGLMode(void);
	virtual void ChangeOpenGLFilter(int num);
	virtual int  GetOpenGLFilter(void);
#endif

	// Sound

	virtual void PostEtStartRecordSound(void);
	virtual void PostEtStopRecordSound(void);
	virtual bool NowRecordingSound(void);
	virtual void ChangeSoundFrequency(int num);
	virtual int  GetSoundFrequencyNum();
	virtual void ChangeSoundLatency(int num);
	virtual int  GetSoundLatencyNum();

#ifdef USE_DATAREC
	// Data Recorder Tape

	virtual void PostEtLoadDataRecMessage(const _TCHAR *file_path);
	virtual void PostEtLoadRecentDataRecMessage(int num);
	virtual void PostEtSaveDataRecMessage(const _TCHAR *file_path);
	virtual void PostEtCloseDataRecMessage(void);
	virtual void PostEtRewindDataRecMessage(void);
	virtual void PostEtFastForwardDataRecMessage(void);
	virtual void PostEtToggleRealModeDataRecMessage(void);
	virtual bool NowRealModeDataRec(void);
	virtual bool IsOpenedLoadDataRecFile(void);
	virtual bool IsOpenedSaveDataRecFile(void);
#endif

#ifdef USE_FD1
	// Floppy disk

	virtual void PostEtOpenFloppyMessage(int drv, const _TCHAR *file_path, int bank, uint32 flags, bool multiopen);
	virtual void PostEtOpenRecentFloppyMessage(int drv, int num);
	virtual void PostEtOpenFloppySelectedVolume(int drv, int bank_num);
	virtual void PostEtCloseFloppyMessage(int drv);
	virtual void PostEtChangeSideFloppyDisk(int drv);
	virtual int  GetSideFloppyDisk(int drv);
	virtual void PostEtToggleWriteProtectFloppyDisk(int drv);
	virtual bool InsertedFloppyDisk(int drv);
	virtual bool WriteProtectedFloppyDisk(int drv);

	virtual bool OpenFloppyDisk(int drv, const _TCHAR* filename, int bank_num, uint32 flags, bool multiopen);
	virtual void CloseFloppyDisk(int drv);

	virtual bool OpenFloppyDiskSelectedVolume(int drv, int bank_num);

	virtual void GetMultiVolumeStr(int num, const _TCHAR *name, _TCHAR *str);

	virtual d88_file_t *GetD88File(int drv);
#endif

#ifdef USE_PRINTER
	// Printer

	virtual void PostEtSavePrinterMessage(int drv, const _TCHAR *file_path);
	virtual int  GetPrinterBufferSize(int drv);
	virtual void PostEtClearPrinterBufferMessage(int drv);
	virtual void PostEtPrintPrinterMessage(int drv);
	virtual void PostEtEnablePrinterDirectMessage(int drv);
	virtual bool IsEnablePrinterDirect(int drv);
#endif

	// Comm

	virtual void PostEtEnableCommServerMessage(int drv);
	virtual bool IsEnableCommServer(int drv);
	virtual void PostEtConnectCommMessage(int drv);
	virtual bool NowConnectingComm(int drv);
	virtual void ToggleCommThroughMode(int drv);
	virtual bool NowCommThroughMode(int drv);
	virtual void ToggleCommBinaryMode(int drv);
	virtual bool NowCommBinaryMode(int drv);
	virtual void SendCommTelnetCommand(int drv, int num);

	// Status

	virtual void PostEtLoadStatusMessage(const _TCHAR *file_path);
	virtual void PostEtSaveStatusMessage(const _TCHAR *file_path, bool sys_pause);

	// Auto Key

	virtual void PostEtLoadAutoKeyMessage(const _TCHAR *file_path);
	virtual void PostEtStartAutoKeyMessage(void);
	virtual void PostEtStopAutoKeyMessage(void);
	virtual bool IsRunningAutoKey(void);

	virtual bool StartAutoKey(void);

	// Record key

	virtual void PostEtLoadRecKeyMessage(const _TCHAR *file_path);
	virtual void PostEtSaveRecKeyMessage(const _TCHAR *file_path, bool sys_pause);
	virtual void StopPlayRecKey(void);
	virtual void StopRecordRecKey(void);
	virtual bool NowPlayingRecKey(void);
	virtual bool NowRecordingRecKey(void);

	// Misc

	virtual void ChangeLedBox(int phase);
	virtual void ToggleLedBox(void);
	virtual void ToggleShowLedBox(void);
	virtual void ToggleInsideLedBox(void);
	virtual int  GetLedBoxPhase(int id);
	virtual bool IsShownLedBox(void);
	virtual bool IsInsidedLedBox(void);
	virtual void ChangeLedBoxPosition(int num);
	virtual void ToggleMessageBoard(void);
	virtual bool IsShownMessageBoard(void);

	virtual void ChangeUseJoypad(int num);
	virtual bool IsEnableJoypad(int num);
	virtual void ToggleEnableLightpen(void);
	virtual bool IsEnableLightpen(void);
	virtual void ToggleUseMouse(void);
	virtual bool IsEnableMouse(void);

	virtual bool GetRecentFileStr(const _TCHAR *file, int num, _TCHAR *str, int trimlen);

#ifdef USE_DEBUGGER
	virtual void OpenDebugger(int num);
	virtual void PostMtOpenDebugger(int num);
	virtual void CloseDebugger();
	virtual void PostMtCloseDebugger();
	virtual bool IsDebuggerOpened(int num);
#endif

	// Drop
	virtual bool OpenFileByExtention(const _TCHAR *file_path);
	virtual bool OpenDroppedFile(void *param);
	//@}

	/// @name utilities
	//@{
	static int CheckSupportedFile(const _TCHAR *file_path);
	static void GetLibVersionString(_TCHAR *str, int max_len = _MAX_PATH, const _TCHAR *sep_str = _T("\n"));
	//@}
};

#endif /* _GUI_BASE_H_ */
