/** @file logging.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2015.02.18 -

	@brief [ logging ]
*/

#include "logging.h"
#include "vm/vm.h"
#include "utility.h"

Logging::Logging()
{
	fplog = NULL;
	mux = new CMutex();
}

Logging::~Logging()
{
	delete mux;
}

void Logging::open_logfile(const _TCHAR *path)
{
#ifdef _LOG_FILE
	// get application path
	char log_path[_MAX_PATH];
	UTILITY::wcs_to_mbs(log_path, path, _MAX_PATH);
	sprintf(&log_path[strlen(log_path)], "%s.log", CONFIG_NAME);

	fplog = fopen(log_path, "w");
#endif
}

void Logging::close_logfile()
{
#ifdef _LOG_FILE
	mux->lock();
	if(fplog) {
		fclose(fplog);
		fplog = NULL;
	}
	mux->unlock();
#endif
}

void Logging::out_logv(int level, const _TCHAR* format, va_list ap)
{
	_TCHAR buffer[1024];
	_TCHAR levelstr[32];

	if (ap) {
#ifndef _UNICODE
#if (defined(_WIN32) && !defined(__MINGW32__))
		UTILITY::vsprintf_utf8(buffer, 1024, format, ap);
#else
		_vstprintf(buffer, format, ap);
#endif
#else
		UTILITY::vstprintf(buffer, 1024, format, ap);
#endif
	} else {
		_tcscpy(buffer, format);
	}

	switch(level) {
	case LOG_WARN:
		_tcscpy(levelstr, _T("WARN"));
		break;
	case LOG_ERROR:
		_tcscpy(levelstr, _T("ERROR"));
		break;
	case LOG_DEBUG:
		_tcscpy(levelstr, _T("DEBUG"));
		break;
	default:
		_tcscpy(levelstr, _T("INFO"));
		break;
	}

	mux->lock();
#ifdef _LOG_CONSOLE
	_ftprintf(stderr, "%s: %s\n", levelstr, buffer);
	fflush(stderr);
#endif

	send_log_message(level, levelstr, buffer);

#ifdef _LOG_FILE
	if(fplog) {
		_ftprintf(fplog, _T("%10d %-5s: %s\n"), SDL_GetTicks(), levelstr, buffer);
		fflush(fplog);
	}
#endif
	mux->unlock();
}

void Logging::out_logf(int level, const _TCHAR* format, ...)
{
	va_list ap;

	va_start(ap, format);
	out_logv(level, format, ap);
	va_end(ap);
}

void Logging::out_log(int level, const _TCHAR* msg)
{
	out_logv(level, msg, NULL);
}

void Logging::out_debugf(const _TCHAR* format, ...)
{
#ifdef _DEBUG_LOG
	va_list ap;

	va_start(ap, format);
	out_logv(LOG_DEBUG, format, ap);
	va_end(ap);
#endif
}

void Logging::out_debug(const _TCHAR* msg)
{
#ifdef _DEBUG_LOG
	out_log(LOG_DEBUG, msg);
#endif
}

#ifdef _UNICODE

void Logging::out_logv(int level, const char* format, va_list ap)
{
	char buffer[1024];
	char levelstr[32];
	_TCHAR wbuffer[1024];
	_TCHAR wlevelstr[32];

	if (ap) {
		vsprintf(buffer, format, ap);
	} else {
		strcpy(buffer, format);
	}

	switch(level) {
	case LOG_WARN:
		strcpy(levelstr, "WARN");
		break;
	case LOG_ERROR:
		strcpy(levelstr, "ERROR");
		break;
	case LOG_DEBUG:
		strcpy(levelstr, "DEBUG");
		break;
	default:
		strcpy(levelstr, "INFO");
		break;
	}

	mux->lock();
#ifdef _LOG_CONSOLE
	fprintf(stderr, "%s: %s\n", levelstr, buffer);
	fflush(stderr);
#endif

	UTILITY::mbs_to_wcs(wbuffer, buffer, 1024);
	UTILITY::mbs_to_wcs(wlevelstr, levelstr, 32);
	send_log_message(level, wlevelstr, wbuffer);

#ifdef _LOG_FILE
	if(fplog) {
		fprintf(fplog, "%10d %-5s: %s\n", SDL_GetTicks(), levelstr, buffer);
		fflush(fplog);
	}
#endif
	mux->unlock();
}

void Logging::out_logf(int level, const char* format, ...)
{
	va_list ap;

	va_start(ap, format);
	out_logv(level, format, ap);
	va_end(ap);
}

void Logging::out_log(int level, const char* msg)
{
	_TCHAR wmsg[1024];
	UTILITY::mbs_to_wcs(wmsg, msg, 1024);
	out_logv(level, wmsg, NULL);
}

void Logging::out_debugf(const char* format, ...)
{
#ifdef _DEBUG_LOG
	va_list ap;

	va_start(ap, format);
	out_logv(LOG_DEBUG, format, ap);
	va_end(ap);
#endif
}

void Logging::out_debug(const char* msg)
{
#ifdef _DEBUG_LOG
	out_log(LOG_DEBUG, msg);
#endif
}

#endif /* _UNICODE */
