/** @file rec_video.cpp

	Skelton for retropc emulator

	@author Sasaji
	@date   2015.05.18 -

	@brief [ record video ]
*/

#include "rec_video.h"
#include "../emu.h"
#include "../sdl_utils.h"

#ifdef USE_REC_VIDEO
#ifdef USE_REC_VIDEO_VFW
#include "vfw/vfw_rec_video.h"
#endif
#ifdef USE_REC_VIDEO_QTKIT
#include "qtkit/qt_rec_video.h"
#endif
#ifdef USE_REC_VIDEO_FFMPEG
#include "ffmpeg/ffm_rec_video.h"
#endif
#ifdef USE_REC_VIDEO_AVKIT
#include "avkit/avk_rec_video.h"
#endif
#ifdef USE_REC_VIDEO_WIN
#include "windows/win_rec_video.h"
#endif
#ifdef USE_REC_VIDEO_COCOA
#include "cocoa/cocoa_rec_video.h"
#endif
#ifdef USE_REC_VIDEO_LIBPNG
#include "libpng/png_rec_video.h"
#endif
#endif /* USE_REC_VIDEO */

REC_VIDEO::REC_VIDEO(EMU *new_emu)
{
	emu = new_emu;
	now_recording = false;
	rec_type = 0;
	rec_fps = 0;
	rec_count = 0;
	rec_denomi = 1;
	memset(rec_path, 0, sizeof(rec_path));
	rec_surface = NULL;

#ifdef USE_REC_VIDEO
#ifdef USE_REC_VIDEO_VFW
	vfwvideo = new VFW_REC_VIDEO(new_emu, this);
#endif
#ifdef USE_REC_VIDEO_QTKIT
	qtvideo = new QT_REC_VIDEO(new_emu, this);
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	ffmvideo = new FFM_REC_VIDEO(new_emu, this);
#endif
#ifdef USE_REC_VIDEO_AVKIT
	avkvideo = new AVK_REC_VIDEO(new_emu, this);
#endif
#ifdef USE_REC_VIDEO_WIN
	winvideo = new WIN_REC_VIDEO(new_emu, this);
#endif
#ifdef USE_REC_VIDEO_COCOA
	cocvideo = new COCOA_REC_VIDEO(new_emu, this);
#endif
#ifdef USE_REC_VIDEO_LIBPNG
	pngvideo = new PNG_REC_VIDEO(new_emu, this);
#endif
#endif /* USE_REC_VIDEO */
}

REC_VIDEO::~REC_VIDEO()
{
#ifdef USE_REC_VIDEO
#ifdef USE_REC_VIDEO_FFMPEG
	delete ffmvideo;
#endif
#ifdef USE_REC_VIDEO_QTKIT
	delete qtvideo;
#endif
#ifdef USE_REC_VIDEO_VFW
	delete vfwvideo;
#endif
#ifdef USE_REC_VIDEO_AVKIT
	delete avkvideo;
#endif
#ifdef USE_REC_VIDEO_WIN
	delete winvideo;
#endif
#ifdef USE_REC_VIDEO_COCOA
	delete cocvideo;
#endif
#ifdef USE_REC_VIDEO_LIBPNG
	delete pngvideo;
#endif
#endif /* USE_REC_VIDEO */

	SDL_UTILS::release_surface(rec_surface);
}

bool REC_VIDEO::IsEnabled(int type)
{
	bool rc = false;
#ifdef USE_REC_VIDEO
	switch(type) {
#ifdef USE_REC_VIDEO_VFW
	case RECORD_VIDEO_TYPE_VFW:
		rc = true;
		break;
#endif
#ifdef USE_REC_VIDEO_QTKIT
	case RECORD_VIDEO_TYPE_QTKIT:
		rc = qtvideo->IsEnabled();
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		rc = ffmvideo->IsEnabled();
		break;
#endif
#ifdef USE_REC_VIDEO_AVKIT
	case RECORD_VIDEO_TYPE_AVKIT:
		rc = avkvideo->IsEnabled();
		break;
#endif
	default:
		break;
	}
#endif
	return rc;
}

const _TCHAR **REC_VIDEO::GetCodecList(int type)
{
	const _TCHAR **list = NULL;
#ifdef USE_REC_VIDEO
	switch(type) {
#ifdef USE_REC_VIDEO_QTKIT
	case RECORD_VIDEO_TYPE_QTKIT:
		list = qtvideo->GetCodecList();
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		list = ffmvideo->GetCodecList();
		break;
#endif
#ifdef USE_REC_VIDEO_AVKIT
	case RECORD_VIDEO_TYPE_AVKIT:
		list = avkvideo->GetCodecList();
		break;
#endif
	default:
		break;
	}
#endif
	return list;
}

const _TCHAR **REC_VIDEO::GetQualityList(int type)
{
	const _TCHAR **list = NULL;
#ifdef USE_REC_VIDEO
	switch(type) {
#ifdef USE_REC_VIDEO_QTKIT
	case RECORD_VIDEO_TYPE_QTKIT:
		list = qtvideo->GetQualityList();
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		list = ffmvideo->GetQualityList();
		break;
#endif
#ifdef USE_REC_VIDEO_AVKIT
	case RECORD_VIDEO_TYPE_AVKIT:
		list = avkvideo->GetQualityList();
		break;
#endif
	default:
		break;
	}
#endif
	return list;
}

bool REC_VIDEO::Start(int type, int fps, SDL_Rect *srcrect, SDL_Surface *srcsurface, bool show_dialog)
{
#ifdef USE_REC_VIDEO
	if (type > 0) {
		rec_type = type;
	}
	if(fps > 0) {
		rec_fps = fps;
		rec_count = 0;
		rec_denomi = 60 / fps;
	}

	if (rec_type <= 0 || rec_fps <= 0) {
		return false;
	}

	CreateFileName(rec_path);

	switch(rec_type) {
#ifdef USE_REC_VIDEO_VFW
	case RECORD_VIDEO_TYPE_VFW:
		rec_surface = SDL_UTILS::release_surface(rec_surface);
		rec_surface = SDL_UTILS::create_surface(srcrect, srcsurface, 2);
		if (rec_surface == NULL) {
			emu->out_logf(LOG_ERROR, _T("Cannot create surface for recording video: %s"), SDL_GetError());
			return false;
		}
		now_recording = vfwvideo->Start(rec_path, rec_fps, srcrect, rec_surface, show_dialog);
		break;
#endif
#ifdef USE_REC_VIDEO_QTKIT
	case RECORD_VIDEO_TYPE_QTKIT: {
		rec_surface = SDL_UTILS::release_surface(rec_surface);
		rec_surface = SDL_UTILS::create_surface(srcrect, srcsurface, 1);
		if (rec_surface == NULL) {
			emu->out_logf(LOG_ERROR, _T("Cannot create surface for recording video: %s"), SDL_GetError());
			return false;
		}
		char crec_path[_MAX_PATH];
		UTILITY::wcs_to_mbs(crec_path, rec_path);
		now_recording = qtvideo->Start(crec_path, rec_fps, rec_surface, show_dialog);
	}	break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		// height set minus, so create the flipped surface
		rec_surface = SDL_UTILS::release_surface(rec_surface);
		rec_surface = SDL_UTILS::create_surface(srcrect, srcsurface, 1);
		if (rec_surface == NULL) {
			emu->out_logf(LOG_ERROR, _T("Cannot create surface for recording video: %s"), SDL_GetError());
			return false;
		}
		now_recording = ffmvideo->Start(rec_path, rec_fps, srcrect, rec_surface, show_dialog);
		break;
#endif
#ifdef USE_REC_VIDEO_AVKIT
	case RECORD_VIDEO_TYPE_AVKIT: {
		rec_surface = SDL_UTILS::release_surface(rec_surface);
		rec_surface = SDL_UTILS::create_surface(srcrect, srcsurface, 2);
		if (rec_surface == NULL) {
			emu->out_logf(LOG_ERROR, _T("Cannot create surface for recording video: %s"), SDL_GetError());
			return false;
		}
		char crec_path[_MAX_PATH];
		UTILITY::wcs_to_mbs(crec_path, rec_path);
		now_recording = avkvideo->Start(crec_path, rec_fps, rec_surface, show_dialog);
	}	break;
#endif
	default:
		break;
	}
#endif
	return now_recording;
}

void REC_VIDEO::Stop()
{
	if (!now_recording) return;
#ifdef USE_REC_VIDEO
	switch(rec_type) {
#ifdef USE_REC_VIDEO_VFW
	case RECORD_VIDEO_TYPE_VFW:
		vfwvideo->Stop();
		break;
#endif
#ifdef USE_REC_VIDEO_QTKIT
	case RECORD_VIDEO_TYPE_QTKIT:
		qtvideo->Stop();
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		ffmvideo->Stop();
		break;
#endif
#ifdef USE_REC_VIDEO_AVKIT
	case RECORD_VIDEO_TYPE_AVKIT:
		avkvideo->Stop();
		break;
#endif
	default:
		break;
	}
	rec_surface = SDL_UTILS::release_surface(rec_surface);
#endif
	now_recording = false;
}

bool REC_VIDEO::Restart()
{
	if (!now_recording) return false;
#ifdef USE_REC_VIDEO
	switch(rec_type) {
#ifdef USE_REC_VIDEO_VFW
	case RECORD_VIDEO_TYPE_VFW:
		now_recording = vfwvideo->Restart();
		break;
#endif
#ifdef USE_REC_VIDEO_QTKIT
	case RECORD_VIDEO_TYPE_QTKIT:
		now_recording = qtvideo->Restart();
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		now_recording = ffmvideo->Restart();
		break;
#endif
#ifdef USE_REC_VIDEO_AVKIT
	case RECORD_VIDEO_TYPE_AVKIT:
		now_recording = avkvideo->Restart();
		break;
#endif
	default:
		break;
	}
#endif
	return now_recording;
}

bool REC_VIDEO::Record(SDL_Rect *srcrect, SDL_Surface *srcsurface, SDL_Rect *dstrect)
{
	if (!now_recording) return false;

#ifdef USE_REC_VIDEO
	// copy surface
	SDL_Rect re = { 0, 0, dstrect->w, dstrect->h };

	switch(rec_type) {
#ifdef USE_REC_VIDEO_VFW
	case RECORD_VIDEO_TYPE_VFW:
		// In OpenGL mode, SDL_BlitSurface does not copy surface....
		if (srcrect->w == dstrect->w && srcrect->h == dstrect->h) {
			SDL_UTILS::copy_surface_flipped(srcsurface, srcrect, rec_surface, &re);
		} else {
			SDL_UTILS::zoomin_surface_flipped(srcsurface, srcrect, rec_surface, &re);
		}
//		emu->mix_ledbox(rec_surface);
		now_recording = vfwvideo->Record(rec_surface);
		break;
#endif
#ifdef USE_REC_VIDEO_QTKIT
	case RECORD_VIDEO_TYPE_QTKIT:
		// In OpenGL mode, SDL_BlitSurface does not copy surface....
		if (srcrect->w == dstrect->w && srcrect->h == dstrect->h) {
#ifndef USE_SDL2
			SDL_UTILS::copy_surface(srcsurface, srcrect, rec_surface, &re);
#else
			SDL_BlitSurface(srcsurface, srcrect, rec_surface, &re);
#endif
		} else {
#ifndef USE_SDL2
			SDL_UTILS::zoomin_surface(srcsurface, srcrect, rec_surface, &re);
#else
			SDL_BlitScaled(srcsurface, srcrect, rec_surface, &re);
#endif
		}
//		emu->mix_ledbox(rec_surface);
		now_recording = qtvideo->Record(rec_surface);
		break;
#endif
#ifdef USE_REC_VIDEO_FFMPEG
	case RECORD_VIDEO_TYPE_FFMPEG:
		// In OpenGL mode, SDL_BlitSurface does not copy surface....
		if (srcrect->w == dstrect->w && srcrect->h == dstrect->h) {
#ifndef USE_SDL2
			SDL_UTILS::copy_surface(srcsurface, srcrect, rec_surface, &re);
#else
			SDL_BlitSurface(srcsurface, srcrect, rec_surface, &re);
#endif
		} else {
#ifndef USE_SDL2
			SDL_UTILS::zoomin_surface(srcsurface, srcrect, rec_surface, &re);
#else
			SDL_BlitScaled(srcsurface, srcrect, rec_surface, &re);
#endif
		}
//		emu->mix_ledbox(rec_surface);
		now_recording = ffmvideo->Record(rec_surface);
		break;
#endif
#ifdef USE_REC_VIDEO_AVKIT
	case RECORD_VIDEO_TYPE_AVKIT:
		// In OpenGL mode, SDL_BlitSurface does not copy surface....
		if (srcrect->w == dstrect->w && srcrect->h == dstrect->h) {
#ifndef USE_SDL2
			SDL_UTILS::copy_surface(srcsurface, srcrect, rec_surface, &re);
#else
			SDL_BlitSurface(srcsurface, srcrect, rec_surface, &re);
#endif
		} else {
#ifndef USE_SDL2
			SDL_UTILS::zoomin_surface(srcsurface, srcrect, rec_surface, &re);
#else
			SDL_BlitScaled(srcsurface, srcrect, rec_surface, &re);
#endif
		}
		//		emu->mix_ledbox(rec_surface);
		now_recording = avkvideo->Record(rec_surface);
		break;
#endif
	default:
		break;
	}
#endif
	return now_recording;
}

bool REC_VIDEO::IsRecordFrame()
{
	if (!now_recording) return false;

	bool rc = (rec_count == 0);
	rec_count = ((rec_count + 1) % rec_denomi);
	return rc;
}

static const _TCHAR *capture_file_ext[] = {
	_T(".bmp"),
	_T(".png"),
	NULL
};

bool REC_VIDEO::Capture(int type, SDL_Rect *srcrect, SDL_Surface *srcsurface, SDL_Rect *dstrect)
{
	if (srcrect == NULL || srcsurface == NULL) {
		return false;
	}

	bool rc = false;
	SDL_Surface *bmpsurface = NULL;
	_TCHAR file_path[_MAX_PATH];
	SDL_Rect re = { 0, 0, dstrect->w, dstrect->h };

	CreateFileName(file_path);

#if !(defined(USE_REC_VIDEO_WIN) || defined(USE_REC_VIDEO_COCOA) || defined(USE_REC_VIDEO_LIBPNG))
	config.capture_type = 0;
#endif
	if (1 < config.capture_type) {
		config.capture_type = 0;
	}

	// add ext
	_tcscat(file_path, capture_file_ext[config.capture_type]);

	// create surface for capture
	switch(config.capture_type) {
	case 1:	// PNG
#if defined(USE_REC_VIDEO_WIN)
		// notice that format always use BGRA
		bmpsurface = SDL_UTILS::create_surface(dstrect, srcsurface, 2);
#elif defined(USE_REC_VIDEO_COCOA)
		// notice that format always use RGBA
		bmpsurface = SDL_UTILS::create_surface(dstrect, srcsurface, 3);
#elif defined(USE_REC_VIDEO_LIBPNG)
		// notice that format always use RGBA
		bmpsurface = SDL_UTILS::create_surface(dstrect, srcsurface, 1);
#endif
		break;
	default: // BMP
		// notice that format is same as screen because SaveBMP cannot out pixeldata
		bmpsurface = SDL_UTILS::create_surface(dstrect, srcsurface);
		break;
	}
	if (bmpsurface == NULL) {
		emu->out_logf(LOG_ERROR, _T("Cannot create screen to capture: %s"), SDL_GetError());
		return false;
	}

	// copy surface
	// In OpenGL mode, SDL_BlitSurface does not copy surface....
	if (srcrect->w == dstrect->w && srcrect->h == dstrect->h) {
#ifndef USE_SDL2
		SDL_UTILS::copy_surface(srcsurface, srcrect, bmpsurface, &re);
#else
		SDL_BlitSurface(srcsurface, srcrect, bmpsurface, &re);
#endif
	} else {
#ifndef USE_SDL2
		SDL_UTILS::zoomin_surface(srcsurface, srcrect, bmpsurface, &re);
#else
		SDL_BlitScaled(srcsurface, srcrect, bmpsurface, &re);
#endif
	}

	switch(config.capture_type) {
	case 1:	// PNG
	{
#if defined(USE_REC_VIDEO_WIN)
		CTchar cfile_path(file_path);
		rc = winvideo->Capture(type, bmpsurface, cfile_path.GetW());
#elif defined(USE_REC_VIDEO_COCOA)
		CTchar cfile_path(file_path);
		rc = cocvideo->Capture(type, bmpsurface, cfile_path.Get());
#elif defined(USE_REC_VIDEO_LIBPNG)
		// fill alpha channel
		SDL_UTILS::fill_surface(bmpsurface, 0xffffffff, bmpsurface->format->Amask);

		CTchar cfile_path(file_path);
		rc = pngvideo->Capture(type, bmpsurface, cfile_path.Get());
#endif
		break;
	}
	default: // BMP
	{
		// save surface
		CTchar cfile_path(file_path);
		rc = SaveBMP(bmpsurface, cfile_path.Get());
		break;
	}
	}
	if (rc) {
		if (type == CAPTURE_SCREEN_TYPE) {
			emu->out_log(LOG_INFO, (CT)_("Screen was saved successfully."));
		}
	} else {
		emu->out_logf(LOG_ERROR, _T("Cannot save screen captured file: %s"), SDL_GetError());
	}

	SDL_UTILS::release_surface(bmpsurface);

	return rc;
}

bool REC_VIDEO::SaveBMP(SDL_Surface *surface, const char *file_name)
{
	return (SDL_SaveBMP(surface, file_name) >= 0);
}

void REC_VIDEO::CreateFileName(_TCHAR *file_path)
{
	int tim[8];
	const _TCHAR *app_path;

	app_path = config.snapshot_path.Length() > 0 ? config.snapshot_path : emu->application_path();
	emu->get_timer(tim);
	_stprintf(file_path, _T("%s%04d-%02d-%02d_%02d-%02d-%02d"), app_path, tim[0], tim[1], tim[2], tim[4], tim[5], tim[6]);
}
